#################################################################
# mmucl.tcl - base mmucl library
# 
# Copyright (C) 1997-1998 Mark Patton
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
###################################################################

namespace eval mmucl {
    namespace export init

    array set Mmucl {
	cfg,reconnect     1
	cfg,actions       1
	cfg,subs          1
	cfg,end_color     grey
	cfg,error_color   {bold red}
	cfg,report_color  {bold green}
	cfg,script_char   /
	cfg,verbatim_char \\
	cfg,timeout       20

	host           ""
	port           ""
	login          {}
	save_cmds      {alias action config sub}
	action_matches {}
	alias_args     {}
    }

    set Mmucl(rc_dir) [file join $env(HOME) .mmucl]
    
    # Map text attribs to ansi codes
    array set Ansi {
	reset 0 bold 1 dim 2 underline 4 blink 5 reverse 7
	
	black 30 red 31 green 32 yellow 33 brown 33 blue 34
	magenta 35 cyan 36 grey 37 gray 37 white 37 default 39
	
	bg_black 40 bg_red 41 bg_green 42 bg_yellow 43 bg_brown 44
	bg_blue 44 bg_magenta 45 bg_cyan 46 bg_grey 47 bg_gray 47
	bg_default 49
    }
}

# mmucl::init --
#
#   Initialize an interp to exec user commands. Procs "MCname"
#   are made available as aliases in the user interp.

proc mmucl::init {} {
    variable Mmucl
    global argv env

    array set Mmucl {
	connect 0
	chars  {}
    }
    
    interp create mmucl
    
    foreach proc [info commands ::mmucl::MC*] {
	interp alias mmucl [string range $proc 11 end] {} $proc
    }
    interp alias mmucl history {} history

    file mkdir [file join $Mmucl(rc_dir) chars]

    set rc [file join $Mmucl(rc_dir) mmucl.rc]
    if {[file exists $rc]} {
	if {[catch {mmucl eval [list source $rc]} error] == 1} {
	    report -error "$rc: $error"
	}
    }
       
    catch {source [file join $Mmucl(rc_dir) .chars]}

    if {[set i [lsearch $argv -exec]] != -1} {
	if {[catch {mmucl eval [lindex $argv [incr i]]} error] == 1} {
	    report -error "command line: $error"
	}
    }

    return
}

# mmucl::display --

#   Display a string sent from the mud.
#   The interface defines it.

#proc mmucl::display {str} {
#    puts -nonewline $str
#    flush stdout
#}

# mmucl::to_ansi --
# 
#   Give str ansi attributes and return it.

proc mmucl::to_ansi {str attribs} {
    variable Ansi

    append new \x1B\[0
    foreach attrib $attribs {
	if {[info exists Ansi($attrib)]} {
	    append new \; $Ansi($attrib)
	}
    }

    return [append new m $str]
}

# mmucl::to_regexp --
#
#   Convert a action or sub pattern to a regexp.

proc mmucl::to_regexp {pattern} {
    regsub -all {[\[\()$^*+.?]} $pattern {\\&} pattern 

    set ansi \x1B\\\[\[0-9\;\]*m

    regsub -all {(^|[^%])%\\\^} $pattern \\1(^|\n|^$ansi|\n$ansi) pattern
    regsub -all {(^|[^%])%w} $pattern "\\1(\[^\r\n\t \]+)" pattern
    regsub -all {(^|[^%])%s} $pattern \\1(\[^\r\n\]*) pattern
    regsub -all {(^|[^%])%c} $pattern {\1(.)} pattern
    regsub -all {(^|[^%])%d} $pattern {\1([0-9.]+)} pattern
    regsub -all {(^|[^*])\\\*} $pattern {\1.*} pattern

    regsub -all {%%([d^wsc])} $pattern {%\1} pattern
    regsub -all {\*\*} $pattern * pattern
   
    # sanity check
    regexp -- $pattern ""

    return $pattern
}

# mmucl::read_mud  --
#
#   Get output from mud, check actions and subs,
#   and display it. The regsub filters out telnet junk.

proc mmucl::read_mud {} {
    variable Mmucl
    variable Action
    variable Sub
    
    set mudout [read $Mmucl(sock)]

    regsub -all \xFF.\[^A-Za-z0-9\]? $mudout "" mudout

    if {$Mmucl(cfg,subs)} {
	foreach sub [array names Sub] {
	    regsub -all -- $Sub($sub) $mudout $Mmucl(spec,$sub) mudout
	}
    }
    
    display $mudout
    
    if {$Mmucl(cfg,actions)} {
	foreach action [array names Action] {
	    if {[regexp -- $Action($action) $mudout 0 1 2 3 4 5 6 7 8 9]} {
		set Mmucl(action_matches) [list $0 $1 $2 $3 $4 $5 $6 $7 $8 $9]
		
		if {[catch {mmucl eval $Mmucl(script,$action)} error] == 1} {
		    report -error "action {$action}: $error"
		}
	    }
	}
    }

    if {[eof $Mmucl(sock)]}  {
	MCdisconnect
    }

    return
}

# mmucl::report --
#
#  Inform the user about an error or other event.

proc mmucl::report {args} {
    variable Mmucl

    set arg1 [lindex $args 0]
    if {$arg1 == "-error"} {
	set error [lindex $args 1]
	MCecho [MCcolor "** " bold  $error $Mmucl(cfg,error_color) " **" bold]
    } else {
	MCecho [MCcolor "** " bold $arg1 $Mmucl(cfg,report_color) " **" bold]
    }

    return
}

# ltrimleft --
#
#   Remove str from the left of each element of a list
#   and return the result.

proc ltrimleft {list str} {
    set n [string length $str]
    set new {}
    
    foreach item $list {
	if {[string first $str $item] == 0} {
	    set item [string range $item $n end]
	}
	lappend new $item
    }
    
    return $new
}

# mmucl::MCgetopt --
#
#   Validate a command that has subcommands and
#   returns the subcommand. 

proc mmucl::MCgetopt {cmdname syntax arglist} {
    array set opt $syntax

    if {[llength $arglist] == 0} {
	return -code error "Usage: $cmdname option ?args ...?"
    }

    set arg1 [lindex $arglist 0]

    if {![info exist opt($arg1)]} {
	return -code error \
	    "bad option $arg1: must be [join [array names opt] {, }]"
    }

    set num_args [lindex $opt($arg1) 0]

    if {[lsearch $num_args [llength $arglist]] == -1} {
	set arglist [concat $arglist [lindex $opt($arg1) 1]]
	
	if {[lsearch $num_args [llength $arglist]] == -1} {
	    return -code error \
		"wrong # args: should be \"$cmdname $arg1 [lindex $opt($arg1) 2]\""
	}
    }
    
    uplevel 1 [list set args $arglist]
    uplevel 1 [list set n [llength $arglist]]

    return $arg1
}

# mmucl::MCecho --
#
#   Write each arg to the display with a space in between them
#   and a newline added to the last.

proc mmucl::MCecho {args} {
    foreach str $args {
	append msg $str " "
    }

    display [append msg \n]

    return
}

# mmucl::MCcolor --

#   Take a list of form {str color ...}, convert each
#   str to color, cat them together, and return the result

proc mmucl::MCcolor {args} {
    variable Mmucl

    foreach {str color} $args {
	append ansi [to_ansi $str $color]
    }
    
    return [append ansi [to_ansi "" $Mmucl(cfg,end_color)]]
}

# mmucl::MCparse --
#
#   If the string starts with "\", it is sent to the mud
#   Starting with "/" indicates a tcl script. Otherwise
#   the string is split up into parts delimited by ";".
#   Parts that are not aliases are sent to the mud.

proc mmucl::MCparse {str} {
    variable Mmucl

    set fchar [string index $str 0]
    if {$fchar == $Mmucl(cfg,verbatim_char)} {
	MCwrite [string range $str 1 end]
    } elseif {$fchar == $Mmucl(cfg,script_char)} {
	if {[catch {set val [mmucl eval [string range $str 1 end]]} error] == 1} {
	    report -error $error
	} elseif {[string length $val]} {
	    MCecho $val
	}
    } elseif {![string length $str]} {
	MCwrite ""
    } else {
	foreach cmd [split $str \;] {
	    regexp {^([^ ]*)(.*)$} $cmd x alias args

	    if {[info exists Mmucl(alias,$alias)]} {
		alias_exec $alias $args
	    } else {
		MCwrite $cmd
	    }
	}
    }

    return
}

# mmucl::MCconnect --
#
#   Attempt to connect to a mud, timing out if it takes
#   longer than cfg,timeout

proc mmucl::MCconnect {args} {
    variable Mmucl

    set n [llength $args]
    
    if {$n == 2 || $n == 3} {
	foreach {host port login} $args break
    } else {
	return -code error \
	    "wrong # args: should be \"connect ?host port ?login??\""
    }

    if {[catch {concat $login}]} {
	return -code error "login must be a list"
    }

    if {$Mmucl(connect)} {
	report -error "already connected"
	return
    }
    
    report "connecting to $host at port $port"
    
    # only attempt one connection at a time.
    catch {after cancel $Mmucl(timeout_id)}
    
    # Set up a timeout.
    set Mmucl(timeout_id) \
	[after [expr {$Mmucl(cfg,timeout) * 1000}] \
			 [list mmucl::connect_timeout $host $port $login]]
    
    if {[catch {set Mmucl(sock) [socket -async $host $port]} error] == 1} {
	after cancel $Mmucl(timeout_id)
	report -error $error
    } else {
	fconfigure $Mmucl(sock) -blocking 0 -translation binary
	
	# Fileevent only gets called once a connect occurs.
	# It sets a variable the after commands check to see if we
	# connect in a specific period of time and times out if we don't.
	
	fileevent $Mmucl(sock) readable \
	    [list mmucl::connect_complete $host $port $login]
    }

    return
}

# mmucl::connect_complete  --
#
#   Complete a connection attempted with MCconnect

proc mmucl::connect_complete {host port login} {
    variable Mmucl

    after cancel $Mmucl(timeout_id)
    array set Mmucl [list connect 1 host $host port $port login $login]
    
    fileevent $Mmucl(sock) readable mmucl::read_mud

    # Useful for profiling
    #fileevent $Mmucl(sock) readable {
    #	mmucl::MCecho [time {mmucl::read_mud}]
    #}
    
    if {[catch {read_mud} error]} {
    	report -error "connection refused"
    	MCdisconnect
    	return
    }

    set host [join [lrange [fconfigure $Mmucl(sock) -peername] 1 end] :]
    report "connected to $host"

    if {$Mmucl(connect) && [llength $login]} {
	eval MCwrite $login
    }

    return
}

# mmucl::connect_timeout --
#
#   Handle a maybe timeout on a connection

proc mmucl::connect_timeout {host port login} {
    variable Mmucl

    if {!$Mmucl(connect)} {
	report "connection timed out"

	catch {close $Mmucl(sock)}
	if {[string compare $Mmucl(cfg,reconnect) 0]} {
	    report "reconnectiong..."
	    after 2000 [list mmucl::MCconnect $host $port $login]
	}
    }

    return
}

# mmucl::MCdisconnect --
#
#   Disconnect from a mud. 

proc mmucl::MCdisconnect {} {
    variable Mmucl

    if {!$Mmucl(connect)} {
	report -error "not connected"
    } else {
	catch {close $Mmucl(sock)}
	set Mmucl(connect) 0
	
	report "connection closed"
    }

    return
}

# mmucl::MCdisconnect --
#
#   Resume the last connection

proc mmucl::MCreconnect {} {
    variable Mmucl

    if {![string length $Mmucl(host)]} {
	report -error "no previous connection"
    } else {
	MCconnect $Mmucl(host) $Mmucl(port) $Mmucl(login)
    }

    return
}

# mmucl::MCconfig --
#
#   Config options. Does some type checking

proc mmucl::MCconfig {args} {
    variable Mmucl
    
    set syntax {
	set    {{2 3} {}  "option ?value?"}
	names  {{2}   {*} "?pattern?"}
	print  {{2}   {*} "?pattern?"}
    }

    set option [MCgetopt config $syntax $args]
    
    if {$option == "set"} {
	set opt [lindex $args 1]

	if {![info exists Mmucl(cfg,$opt)]} {
	    report -error "bad option $opt"
	} elseif {$n == 2} {
	    return $Mmucl(cfg,$opt)
	} else {
	    set val [lindex $args 2]

	    if {[string match *_char $opt]} {
		if {[string length $val] != 1} {
		    return -code error "value must be a char"
		}
	    } elseif {[string match *_color $opt]} {
		if {[catch {concat $val}]} {
		    return -code error "value must be a list"
		}
	    } elseif {![regexp {^[0-9]+$} $val]} {
		return -code error "value must be a number"
	    }

	    set Mmucl(cfg,$opt) $val
	}
    } elseif {$option == "names"} {	
	return [ltrimleft [array names Mmucl cfg,[lindex $args 1]] cfg,]
    } else {
	foreach name [MCconfig names [lindex $args 1]] {
	    MCecho $name: $Mmucl(cfg,$name)
	}
    }

    return
}

# mmucl::MCwrite --
#
#   Write each arg to the mud
 
proc mmucl::MCwrite {args} {
    variable Mmucl
    
    if {!$Mmucl(connect)} {
	report -error "not connected"
    } else {
	foreach str $args {
	    puts $Mmucl(sock) $str
	}
	
	flush $Mmucl(sock)
    }
    
    return
}

# mmucl::MCalias --
#
#   Add, delete, and inspect aliases.

proc mmucl::MCalias {args} {
    variable Mmucl

    set syntax {
	set    {{2 3} {}  "name ?script?"}
	names  {{2}   {*} "?pattern?"}
	print  {{2}   {*} "?pattern?"}
	delete {{2}   {}  "pattern"}
	exec   {{3}  {{}} "name ?arg_list?"}
    }

    set option [MCgetopt alias $syntax $args]

    if {$option == "set"} {
	set name [lindex $args 1]
	
	if {$n == 3} {
	    set Mmucl(alias,$name) [lindex $args 2]
	} elseif {![info exists Mmucl(alias,$name)]} {
	    report -error "no such alias"
	} else {
	    return $Mmucl(alias,$name)
	}
    } elseif {$option == "names"} {	
	return [ltrimleft [array names Mmucl alias,[lindex $args 1]] alias,]
    } elseif {$option == "delete"} {
	foreach name [array names Mmucl alias,[lindex $args 1]] {
	    unset Mmucl($name)
	}
    } elseif {$option == "print"} {
	foreach name [MCalias names [lindex $args 1]] {
	    MCecho "\"$name\" set to {$Mmucl(alias,$name)}"
	}
    } else {
	alias_exec [lindex $args 1] [lindex $args 2]
    }

    return
}

# mmucl::exec_alias --
#
#   Execute alias name

proc mmucl::alias_exec {name alias_args}  {
    variable Mmucl

    if {![info exists Mmucl(alias,$name)]} {
	report -error "no such alias"
    } elseif {[catch {set Mmucl(alias_args) [concat $alias_args]} error] == 1} {
	report -error "alias args: $error"
    } elseif {[catch {mmucl eval $Mmucl(alias,$name)} error] == 1} {
	report -error "alias \"$name\": $error"
    }

    return
}

# mmucl::MC%
#
#   Access args to an alias.

proc mmucl::MC% {{i 0}} {
    variable Mmucl

    if {[catch {incr i -1} error]} {
	return -code error "arg must be a number"
    } elseif {$i == -1} {
	return $Mmucl(alias_args)
    } else {
	return [lindex $Mmucl(alias_args) $i]
    }
}

# mmucl::MCaction --
#
#   Add or display actions

proc mmucl::MCaction {args} {
    variable Mmucl
    variable Action

    set syntax {
	set    {{2 3} {}  "pattern ?script?"}
	names  {{2}   {*} "?pattern?"}
	print  {{2}   {*} "?pattern?"}
	delete {{2}   {}  "pattern"}
    }

    set option [MCgetopt action $syntax $args]    

    if {$option == "set"} {	    
	set pattern [lindex $args 1]
	
	if {$n == 3} {
	    set Action($pattern) [to_regexp $pattern]
	    set Mmucl(script,$pattern) [lindex $args 2]
	} elseif {![info exists Mmucl(script,$pattern)]} {
	    report -error "no such action"
	} else {
	    return $Mmucl(script,$pattern)
	}
    } elseif {$option == "names"} {
	return [array names Action [lindex $args 1]]
    } elseif {$option == "delete"} {
	foreach name [array names Action [lindex $args 1]] {
	    unset Mmucl(script,$name) Action($name)
	}
    } else {
	foreach name [array names Action [lindex $args 1]] {
	    MCecho "{$name} triggers {$Mmucl(script,$name)}"
	}
    }

    return
}

proc mmucl::MC@ {{match 0}} {
    variable Mmucl

    if {[catch {incr match -1} error]} {
	return -code error "arg must be a number"
    }
    
    return [lindex $Mmucl(action_matches) [incr match]]
}

proc mmucl::MCsub {args} {
    variable Mmucl
    variable Sub

    set syntax {
	set    {{2 3} {}  "pattern script"}
	names  {{2}   {*} "?pattern?"}
	print  {{2}   {*} "?pattern?"}
	delete {{2}   {}  "pattern"}
    }

    set option [MCgetopt sub $syntax $args]
    
    if {$option == "set"} {
	set pattern [lindex $args 1]
	
	if {$n == 3} {
	    set Sub($pattern) [to_regexp $pattern]
	    set Mmucl(spec,$pattern) [lindex $args 2]
	} elseif {![info exists Sub($pattern)]} {
	    return -code error "no such sub"
	} else {
	    return $Sub($pattern)
	}
    } elseif {$option == "names"} {
	return [array names Sub [lindex $args 1]]
    } elseif {$option == "delete"} {
	foreach name [MCsub names [lindex $args 1]] {
	    unset Sub($name) Mmucl(spec,$name)
	}
    } else {
	foreach name [array names Sub [lindex $args 1]] {
	    MCecho "{$name} changed to {$Mmucl(spec,$name)}"
	}
    }

    return
}


# mmucl::MCchar --
#

proc mmucl::MCchar {args} {
    variable Mmucl

    set syntax {
	set    {{2 6} {{}} "name ?mud host port ?login??"}
	names  {{2}   {*}  "?pattern?"}
	print  {{2}   {*}  "?pattern?"}
	delete {{2}   {}   "pattern"}
	load   {{2}   {}   "char"}
    }

    set option [MCgetopt char $syntax $args]

    array set char $Mmucl(chars)

    if {$option == "set"} {
	if {$n == 6} {
	    # create init file
	    catch {close [open [file join $Mmucl(rc_dir) char \
				    [lindex $args 1].[lindex $args 2]] a+]}

	    set char([lindex $args 1]) [lrange $args 2 end]
	} elseif {![info exists char([lindex $args 1])]} {
	    report -error "no such char"
	} else {
	    return $char([lindex $args 1])
	}
    } elseif {$option == "names"} {	
	return [array names char [lindex $args 1]]
    } elseif {$option == "delete"} {
	foreach name [array names char [lindex $args 1]] {
	    unset char($name)
	}
    } elseif {$option == "load"} {
	set name [lindex $args 1]

	if {![info exists char($name)]} {
	    return -code error "no such char"
	}
	
	foreach {mud host port login} $char($name) {}
	
	set init [file join $Mmucl(rc_dir) chars $name.$mud]
	if {[file exists $init]} {
	    mmucl eval [list source $init]
	}
	
	MCconnect $host $port $login
    } else {
	foreach name [array names char [lindex $args 1]] {
	    MCecho $name
	    MCecho "  Mud:   [lindex $char($name) 0]"
	    MCecho "  Host:  [lindex $char($name) 1]:[lindex $char($name) 2]"
	    if {[llength [lindex $char($name) 3]]} {
		MCecho "  Login: [lindex $char($name) 3]"
	    }
	}
    }
    
    set Mmucl(chars) [array get char]
    
    return
}

proc mmucl::MCsave {filename} {
    variable Mmucl
    
    if {[catch {set fd [open $filename w]} error]} {
	report -error $error
	return
    }
    
    foreach cmd $Mmucl(save_cmds) {
	foreach name [MC$cmd names] {
	    puts $fd [list $cmd set $name [MC$cmd set $name]]
	}
	puts $fd ""
    }
	
    flush $fd
    close $fd
    
    return
}

# mmucl::MCtextin --
#
#   Read a file directly into the mud.

proc mmucl::MCtextin {filename} {
    if {[catch {set fd [open $filename r]} error]} {
	report -error $error
	return
    }
    
    MCwrite [read $fd]
    close $fd
    
    return
}

# mmucl::MCexit --
#
#   Exit the app and write .chars

proc mmucl::MCexit {} {
    variable Mmucl
    
    set fd [open [file join $Mmucl(rc_dir) .chars] w]
    puts $fd [list set mmucl::Mmucl(chars) $Mmucl(chars)]
    flush $fd
    close $fd

    exit
}
